import json
import logging

import click

from ...codebase import providers
from .. import helpers
from ..helpers import *
from ..http_client import client


@click.command(
    name="grab", help="Stream the stream links to the stdout stream for external usage."
)
@click.argument("query", required=True)
@click.option(
    "-r",
    "--range",
    help="Select ranges of anime.",
    required=False,
    default=":",
    type=str,
)
@click.option(
    "-f",
    "--file",
    help="File to write all the grabbed content to.",
    required=False,
    default="",
    show_default=False,
    type=str,
)
@click.option(
    "--auto",
    is_flag=True,
    default=False,
    help="Select the first given index without asking for prompts.",
)
@click.option(
    "-i",
    "--index",
    required=False,
    default=0,
    show_default=False,
    type=int,
    help="Index for the auto flag.",
)
@click.option(
    "--log-file",
    help="Set a log file to log everything to.",
    required=False,
)
@click.option(
    "-ll", "--log-level", help="Set the integer log level.", type=int, default=20
)
@bannerify
def animdl_grab(query, file, auto, index, log_level, **kwargs):

    r = kwargs.get("range")

    session = client
    logger = logging.getLogger("grabber")
    anime, provider = process_query(session, query, logger, auto=auto, auto_index=index)
    if not anime:
        return
    logger.name = "{}/{}".format(provider, logger.name)
    logger.info("Initializing grabbing session.")

    if file:
        collected_streams = []
        file += ".json" if not file.endswith(".json") else ""

    for stream_url_caller, episode in providers.get_appropriate(
        session, anime.get("anime_url"), check=get_check(r)
    ):
        stream_url = list(helpers.ensure_extraction(session, stream_url_caller))

        if file:
            collected_streams.append({"episode": episode, "streams": stream_url})

        if file:
            logger.info("{} => {!r}".format("E%02d" % episode, file))
            try:
                with open(file, "w") as json_file_writer:
                    json.dump(collected_streams, json_file_writer, indent=4)
            except WindowsError:
                logger.error(
                    "Failed to attempt I/O on the file at the moment; the unwritten value(s) will be written in the next I/O."
                )
        else:
            to_stdout(
                json.dumps({"episode": episode, "streams": stream_url}),
                ("E%02d" % episode) if log_level <= 20 else "",
            )
    logger.info("Grabbing session complete.")



import subprocess
import sys

from click import prompt

from ...config import FZF_EXECUTABLE, FZF_OPTS, FZF_STATE


def default_prompt(
    logger,
    components,
    *,
    processor=None,
    component_name="result",
    fallback=None,
    error_message=None,
    stdout_processor=None,
    escape_output=False,
):

    if processor is None:
        components = list(components)
    else:
        components = list(processor(component) for component in components)

    if not components:
        if error_message is not None:
            logger.critical(error_message)
        return fallback

    for component_id, component in enumerate(components, 1):
        out = stdout_processor(component) if stdout_processor else component

        if escape_output:
            out = repr(out)[1:-1]

        logger.info("{:02d}: {}".format(component_id, out))

    if len(component) == 1:
        logger.debug(
            "The prompt list has only one {}. Returning it.".format(component_name)
        )
        return components[0]

    user_selection = (
        prompt(
            "Select the {} using index".format(component_name),
            default=1,
            type=int,
            show_default=True,
        )
        - 1
    )

    return components[user_selection % len(components)]


def fzf_prompt(
    logger,
    components,
    *,
    processor=None,
    component_name="result",
    fallback=None,
    error_message=None,
    stdout_processor=None,
    escape_output=False,
):
    if processor is None:
        components = list(components)
    else:
        components = list(processor(component) for component in components)

    stdout_mapout = {}

    for component_output, component in zip(
        (stdout_processor(component) for component in components)
        if stdout_processor
        else components,
        components,
    ):

        if not isinstance(component_output, str):
            raise TypeError(
                "The stdout_processor must return a string and not {!r}.".format(
                    type(component_output)
                )
            )

        if escape_output:
            component_output = repr(component_output)[1:-1]

        if component_output in stdout_mapout:
            component_output += " (apparent duplicate {})".format(component_name)

        stdout_mapout[component_output] = component

    if not stdout_mapout:
        if error_message is not None:
            logger.critical(error_message)
        return fallback

    fzf_args = [
        FZF_EXECUTABLE,
        "--header={}".format(
            f"Select the {component_name} (automatically selects the top {component_name})"
        ),
    ] + FZF_OPTS

    process = subprocess.Popen(
        fzf_args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )

    selection, _ = process.communicate(b"\n".join(map(str.encode, stdout_mapout)))

    if process.returncode in [1, 130]:
        return components[0]

    return stdout_mapout[selection.decode()[:-1]]


def get_prompt_manager(*, fallback=default_prompt):

    if not sys.stdout.isatty():
        return fallback

    if FZF_STATE:
        return fzf_prompt

    return fallback



import regex
import logging

STREAMTAPE_REGEX = regex.compile(r"'robotlink'\)\.innerHTML = '(.+?)'\+ \('xcd(.+?)'\)")


def extract(session, url, **opts):
    """
    A safe extraction for Streamtape.
    """
    logger = logging.getLogger("streamtape-extractor")
    streamtape_embed_page = session.get(url)
    regex_match = STREAMTAPE_REGEX.search(streamtape_embed_page.text)
    if not regex_match:
        logger.warning(
            "Could not find stream links. {}".format(
                "The file was deleted."
                if streamtape_embed_page.status_code == 404
                else "Failed to extract from: {}".format(url)
            )
        )
        return []

    content_get_uri = "https:{}".format("".join(regex_match.groups()))

    streamtape_redirect = session.get(content_get_uri, follow_redirects=False)
    return [{"stream_url": streamtape_redirect.headers.get("location")}]



import json
from functools import partial

import lxml.html as htmlparser
import regex

from ....config import NINEANIME
from ...helper import construct_site_based_regex
from .decipher import decrypt_url, encrypt_url

REGEX = construct_site_based_regex(
    NINEANIME, extra_regex=r"/watch/[^&?/]+\.(?P<slug>[^&?/]+)"
)
TITLES_REGEX = regex.compile(r'<h1 itemprop="name" class="title" .+?>(.+?)</h1>')

SOURCES = {
    "41": "vidstream",
    "28": "mycloud",
    "35": "mp4upload",
    "40": "streamtape",
    "43": "videovard",
}


def fetch_episode(session, data_sources):
    for server_id, data_hash in data_sources.items():
        response = session.get(
            NINEANIME + "ajax/anime/episode", params={"id": data_hash}
        )
        if response.status_code >= 400:
            continue

        yield {
            "stream_url": decrypt_url(response.json().get("url")),
            "further_extraction": (
                SOURCES.get(server_id),
                {"headers": {"referer": NINEANIME}},
            ),
        }


def fetcher(session, url, check, match):
    slug = match.group("slug")

    for episode in htmlparser.fromstring(
        session.get(
            NINEANIME + "ajax/anime/servers",
            params={"id": slug, "vrf": encrypt_url(slug)},
        )
        .json()
        .get("html")
    ).cssselect("a[title][data-sources][data-base]"):
        number = int(episode.get("data-base", 0))
        if check(number):
            yield partial(
                lambda data_sources: fetch_episode(session, data_sources),
                data_sources=json.loads(episode.get("data-sources")),
            ), number


def metadata_fetcher(session, url, match):

    response = session.get(url).text

    return {"titles": TITLES_REGEX.findall(response)}



from click import prompt

from ...codebase.providers import get_provider
from ...config import DEFAULT_PROVIDER
from .searcher import get_searcher

from .prompts import get_prompt_manager


def prompt_user(logger, anime_list_genexp, provider):

    manager = get_prompt_manager()

    return manager(
        logger,
        anime_list_genexp,
        processor=lambda component: (component, provider),
        component_name="search result",
        fallback=({}, None),
        error_message=f"Failed to find anything of that query on {provider!r}. Try searching on other providers.",
        stdout_processor=lambda component: f"{component[0]['name']} / {component[0]['anime_url']}",
    )


def process_query(
    session, query: str, logger, *, provider=DEFAULT_PROVIDER, auto=False, auto_index=1
):

    _, module, provider_name = get_provider(query, raise_on_failure=False)

    if module:
        return {"anime_url": query}, provider_name

    *provider_name, custom_query = query.split(":", 1)

    searcher = get_searcher(":".join(provider_name))

    if not searcher:
        searcher, custom_query = get_searcher(provider), query

    genexp = searcher(session, custom_query)

    if not auto:
        return prompt_user(logger, genexp, searcher.provider)
    return list(genexp)[auto_index - 1], searcher.provider



import contextlib
import logging
import os
import shutil
import subprocess
from collections import defaultdict

import regex
from tqdm import tqdm

executable = "ffmpeg"


def has_ffmpeg():
    return bool(shutil.which(executable))


FFMPEG_EXTENSIONS = ["mpd", "m3u8", "m3u"]


def parse_ffmpeg_duration(dt: str) -> float:
    """
    Converts ffmpeg duration to seconds.

    Returns
    ---

    `float`
    """
    hour, minute, seconds = (float(_) for _ in dt.split(":"))
    return hour * (60**2) + minute * 60 + seconds


def iter_audio(stderr):
    """
    Goes over the audio part of the ffmpeg output and gets the mapping index and
    the frequency.

    Returns
    ---

    `Generator[tuple(str, int)]`

    """

    def it():
        """
        A generator, that is made for sorting and sending to another generator.
        """
        for match in regex.finditer(b"Stream #(\d+):(\d+): Audio:.+ (\d+) Hz", stderr):
            program, stream_id, freq = (_.decode() for _ in match.groups())
            yield "{}:a:{}".format(program, stream_id), int(freq)

    yield from sorted(it(), key=lambda x: x[1], reverse=True)


def analyze_stream(logger: logging.Logger, url: str, headers: dict):
    """
    Converts the output of `ffmpeg -i $URL` to a partial stream info default dict.

    In logging level DEBUG, it shows the ffmpeg output.

    Returns
    ---

    `collections.defaultdict`

    """
    info = defaultdict(lambda: defaultdict(lambda: defaultdict(defaultdict)))

    args = [executable, "-hide_banner"]

    if headers:
        args.extend(
            ("-headers", "\r\n".join("{}:{}".format(k, v) for k, v in headers.items()))
        )

    args.extend(("-i", url))

    logger.debug("Calling PIPE child process for ffmpeg: {}".format(args))
    process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    stderr = b"".join(iter(process.stdout))

    duration = regex.search(b"Duration: ((?:\d+:)+\d+)", stderr)
    if duration:
        info["duration"] = parse_ffmpeg_duration(duration.group(1).decode())

    audio = [*iter_audio(stderr)]

    for match in regex.finditer(b"Stream #(\d+):(\d+): Video: .+x(\d+)", stderr):
        program, stream_index, resolution = (int(_.decode()) for _ in match.groups())
        info["streams"][program][stream_index]["quality"] = resolution
        info["streams"][program][stream_index]["audio"] = audio

    return info


def iter_quality(quality_dict):
    """
    Iterates over the quality dict returned by `analyze_stream`.
    """
    for programs, streams in quality_dict.get("streams", {}).items():
        for stream, stream_info in streams.items():
            yield "{}:v:{}".format(programs, stream), stream_info.get("quality") or 0, (
                stream_info.get("audio") or [None, 0]
            )[0][0]


def get_last(iterable):
    """
    Gets the last element from the iterable. Pretty self-explanatory.
    """
    expansion = [*iterable]
    if expansion:
        return expansion[-1]


def ffmpeg_to_tqdm(
    logger: logging.Logger, process: subprocess.Popen, duration: int, outfile_name: str
) -> subprocess.CompletedProcess:
    """
    tqdm wrapper for a ffmpeg process.

    Takes a logger `logger`, the ffmpeg child process `process`, duration of stream
    `duration` and the output file's name `outfile_name`

    This uses the simple concept, stream reading using `iter`, after which it takes
    the current time, converts it into seconds and shows the full progress bar.

    In logging level DEBUG, it shows the ffmpeg output.

    Returns
    ---

    `subprocess.Popen` but completed

    """
    progress_bar = tqdm(
        desc="HLS, FFMPEG / GET {}.mkv".format(outfile_name),
        total=duration,
        unit="segment",
    )
    previous_span = 0

    for stream in process.stdout:
        logger.debug("[ffmpeg] {}".format(stream.decode().strip()))
        current = get_last(regex.finditer(b"\stime=((?:\d+:)+\d+)", stream))
        if current:
            in_seconds = (
                parse_ffmpeg_duration(current.group(1).decode()) - previous_span
            )
            previous_span += in_seconds
            progress_bar.update(in_seconds)

    progress_bar.close()
    return process


def ffmpeg_download(
    url: str,
    headers: dict,
    outfile_name: str,
    content_dir,
    preferred_quality=1080,
    log_level=20,
    **opts
) -> int:
    """
    Downloads content using ffmpeg and optionally uses tqdm to wrap the progress
    bar.

    Initally, it fetches content information for the stream using `analyze_stream`.

    Then after, it selects the quality preferred by the user and maps it to the best
    audio. The stream is then passed to tqdm if the logging level is less than INFO.
    If the logging level is greater than INFO, it simply runs the command and waits.

    In logging level DEBUG, it shows the ffmpeg output.

    Returns
    ---

    `int` The ffmpeg child process' return code.

    """

    logger = logging.getLogger("ffmpeg-hls-download[{}.mkv]".format(outfile_name))
    logger.debug("Using ffmpeg to download content.")

    stream_info = analyze_stream(logger, url, headers)

    file = content_dir / ("{}.mkv".format(outfile_name))

    with contextlib.suppress(FileNotFoundError, OSError):
        os.remove(file)

    args = [executable, "-hide_banner"]

    if headers:
        args.extend(
            ("-headers", "\r\n".join("{}:{}".format(k, v) for k, v in headers.items()))
        )

    args.extend(("-i", url, "-c", "copy", file.as_posix()))

    for video, quality, audio in filter(
        lambda x: x[1] <= preferred_quality,
        sorted(iter_quality(stream_info), key=lambda x: x[1], reverse=True),
    ):
        if quality < preferred_quality:
            logger.warning(
                "Could not find the stream of desired quality {}, currently downloading {}.".format(
                    preferred_quality, quality or "an unknown quality"
                )
            )

        child = subprocess.Popen(
            args + ["-map", video, "-map", audio],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        if log_level > 20:
            return child.wait()

        return ffmpeg_to_tqdm(
            logger,
            child,
            duration=stream_info.get("duration"),
            outfile_name=outfile_name,
        ).returncode


def merge_subtitles(video_path, out_path, subtitle_paths, log_level=20):

    args = [
        executable,
        "-hide_banner",
        "-i",
        video_path,
        "-c:v",
        "copy",
        out_path,
        "-y",
    ]

    for subtitle in subtitle_paths:
        args.extend(("-i", subtitle))

    child = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    if log_level > 20:
        child.wait()
        return child.returncode

    for _ in iter(child.stdout):
        print("[ffmpeg/submerge] {}".format(_.decode("utf-8").strip()), end="\r")
    child.wait()

    return child.returncode



import logging
from collections import defaultdict

import click

from ...codebase import providers
from ...config import DEFAULT_PLAYER, QUALITY, DISCORD_PRESENCE
from .. import exit_codes, helpers, http_client
from ..helpers.intelliq import filter_quality

if DISCORD_PRESENCE:
    try:
        from ..helpers.rpc import set_streaming_episode
    except ImportError:
        raise ImportError(
            "Discord RPC was set to be enabled but `pypresence` is not installed, install it with `pip install pypresence`."
        )
    except RuntimeError:
        DISCORD_PRESENCE = False


def quality_prompt(log_level, logger, stream_list):
    title_dict = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    for n, anime in enumerate(stream_list, 1):
        title_dict[anime.get("title") or "Uncategorized"][
            anime.get("quality") or "No specific quality mentioned"
        ][
            "Soft subtitles (Subtitles are not forced.)"
            if anime.get("subtitle")
            else "Hard subtitles (Subtitles are forced.)"
        ].append(
            "{:02d}: {}".format(n, helpers.stream_judiciary(anime.get("stream_url")))
        )

    for category, qualities in title_dict.items():
        logger.info("\x1b[91m▽ {}\x1b[39m".format(category))
        for quality, subtitles in qualities.items():
            logger.info("\x1b[96m▽▽ {}\x1b[39m".format(quality))
            for subtitle, animes in subtitles.items():
                logger.info("\x1b[95m▽▽▽ {}\x1b[39m".format(subtitle))
                for anime in animes:
                    logger.info(anime)

    return stream_list[
        (
            helpers.ask(
                log_level,
                text="Select above, using the stream index",
                show_default=True,
                default=1,
                type=int,
            )
            - 1
        )
        % len(stream_list)
    ]


@click.command(name="stream", help="Stream your favorite anime by query.")
@click.argument("query", required=True)
@click.option(
    "-r",
    "--range",
    help="Select ranges of anime.",
    required=False,
    default=":",
    type=str,
)
@click.option(
    "-s",
    "--special",
    help="Special range selection.",
    required=False,
    default="",
    type=str,
)
@click.option(
    "--player-opts",
    help="Arguments that are to be passed to the player call.",
    required=False,
)
@click.option(
    "-q",
    "--quality",
    help="Use quality strings.",
    required=False,
    default=QUALITY,
)
@click.option(
    "-p",
    "--player",
    help="Select which player to play from.",
    required=False,
    default=DEFAULT_PLAYER,
    show_default=True,
    show_choices=True,
    type=click.Choice(
        ("mpv", "vlc", "iina", "celluloid", "ffplay", "android"), case_sensitive=False
    ),
)
@click.option(
    "--auto",
    is_flag=True,
    default=False,
    help="Select the first given index without asking for prompts.",
)
@click.option(
    "-i",
    "--index",
    required=False,
    default=0,
    show_default=False,
    type=int,
    help="Index for the auto flag.",
)
@click.option(
    "--log-file",
    help="Set a log file to log everything to.",
    required=False,
)
@click.option(
    "-ll", "--log-level", help="Set the integer log level.", type=int, default=20
)
@helpers.bannerify
def animdl_stream(
    query, special, player_opts, quality, player, auto, index, log_level, **kwargs
):
    """
    Streamer call for animdl streaming session.
    """
    r = kwargs.get("range")

    session = http_client.client
    logger = logging.getLogger("streamer")
    streamer = helpers.handle_streamer(
        click.parser.split_arg_string(player_opts or "") or [], **{player: True}
    )

    if streamer is False:
        logger.critical(
            "Streaming failed due to selection of a unsupported streamer; please configure the streamer in the config to use it."
        )
        raise SystemExit(exit_codes.STREAMER_CONFIGURATION_REQUIRED)

    anime, provider = helpers.process_query(
        session, query, logger, auto=auto, auto_index=index
    )

    if not anime:
        logger.critical("Searcher returned no anime to stream, failed to stream.")
        raise SystemExit(exit_codes.NO_CONTENT_FOUND)

    logger.name = "{}/{}".format(provider, logger.name)
    logger.debug("Will scrape from {}".format(anime))
    logger.info("Now initiating your stream session")

    match, provider_module, _ = providers.get_provider(
        providers.append_protocol(anime.get("anime_url"))
    )

    streams = list(
        provider_module.fetcher(
            session, anime.get("anime_url"), helpers.get_check(r), match
        )
    )

    if special:
        streams = list(helpers.special_parser(streams, special))

    if "name" not in anime:
        anime["name"] = (
            provider_module.metadata_fetcher(session, anime.get("anime_url"), match)[
                "titles"
            ]
            or [None]
        )[0] or ""

    content_title = anime["name"]

    total = len(streams)
    for count, (stream_urls_caller, episode_number) in enumerate(streams, 1):

        playing = True
        while playing:

            window_title = content_title + ": Episode {}".format(episode_number)

            stream_urls = filter_quality(
                list(helpers.ensure_extraction(session, stream_urls_caller)), quality
            )

            if not stream_urls:
                logger.warning(
                    "There were no stream links available at the moment. Ignoring {!r}, retry using a different provider.".format(
                        window_title
                    )
                )
                playing = False
                continue

            selection = (
                quality_prompt(log_level, logger, stream_urls)
                if len(stream_urls) > 1
                else stream_urls[0]
            )

            logger.debug("Calling streamer for {!r}".format(stream_urls))

            headers = selection.get("headers", {})
            _ = headers.pop("ssl_verification", True)

            logger.info(
                "Streaming {!r}, [{:d}/{:d}, {:d} remaining]".format(
                    window_title, count, total, total - count
                )
            )

            if "title" in selection:
                window_title += " ({})".format(selection["title"])

            player_process = streamer(
                selection.get("stream_url"),
                headers=headers,
                content_title=window_title,
                subtitles=selection.get("subtitle", []),
            )
            if DISCORD_PRESENCE:
                set_streaming_episode(session, content_title, episode_number)

            player_process.wait()

            if player_process.returncode:
                logger.warning(
                    "Detected a non-zero return code: {:d}.".format(
                        player_process.returncode
                    )
                )
                playing = (
                    False
                    if log_level > 20
                    else click.confirm(
                        "Retry playback for {!r}?".format(window_title),
                        show_default=True,
                        default=False,
                    )
                )
                continue

            playing = (
                False
                if log_level > 10
                else click.confirm(
                    "Replay {!r}?".format(window_title), show_default=True, default=True
                )
            )



import regex

from ....config import ANIMEPAHE
from ...helper import construct_site_based_regex
from .inner import get_stream_url_from_kwik

from functools import partial, lru_cache

REGEX = construct_site_based_regex(ANIMEPAHE, extra_regex=r"/(?:anime|play)/([^?&/]+)")


API_URL = ANIMEPAHE + "api"
SITE_URL = ANIMEPAHE

PLAYER_RE = construct_site_based_regex(ANIMEPAHE, extra_regex=r"/play/([^?&/]+)")
ID_RE = regex.compile(r'let id = "(.+?)"')
KWIK_RE = regex.compile(r"Plyr\|querySelector\|document\|([^\\']+)")

TITLES_REGEX = regex.compile(r"<h1>(.+?)</h1>")


@lru_cache()
def get_session_page(session, page, release_id):
    return session.get(
        API_URL,
        params={"m": "release", "id": release_id, "sort": "episode_desc", "page": page},
    ).json()


def get_m3u8_from_kwik(session, kwik_url):
    """
    (Unused at the moment!)
    """
    kwik_page = session.get(kwik_url, headers={"referer": SITE_URL})
    match = KWIK_RE.search(kwik_page.text)
    if match:
        return "{10}://{9}-{8}.{7}.{6}.{5}/{4}/{3}/{2}/{1}.{0}".format(
            *match.group(1).split("|")
        )
    raise Exception(
        "Session fetch failure; please recheck and/or retry fetching anime URLs again. If this problem persists, please make an issue immediately."
    )


def get_stream_url(session, release_id, stream_session):

    stream_url_data = session.get(
        API_URL,
        params={"m": "links", "id": release_id, "session": stream_session, "p": "kwik"},
    )
    content = stream_url_data.json().get("data", [])

    for d in content:
        for quality, data in d.items():
            yield {
                "quality": quality,
                "headers": {"referer": data.get("kwik")},
                "stream_url": get_m3u8_from_kwik(session, data.get("kwik")),
            }


def get_stream_urls_from_page(session, release_id, page, check):
    data = get_session_page(session, page, release_id).get("data")
    for content in reversed(data):
        if check(content.get("episode", 0)):
            yield partial(
                lambda session, release_id, content: (
                    [*get_stream_url(session, release_id, content.get("session"))]
                ),
                session,
                release_id,
                content,
            ), content.get("episode", 0)


def predict_pages(total, check):
    """
    A calculative function to minimize API calls.
    """
    for x in range(1, total + 1):
        if check(x):
            yield (total - x) // 30 + 1


def page_minimization(page_generator):
    return sorted(list(dict.fromkeys(page_generator)), reverse=True)


def bypass_ddos_guard(session):
    js_bypass_uri = regex.search(
        r"'(.*?)'", session.get("https://check.ddos-guard.net/check.js").text
    ).group(1)
    session.cookies.update(session.get(ANIMEPAHE + js_bypass_uri).cookies)


def fetcher(session, url, check, match):
    player_match = PLAYER_RE.search(url)

    if player_match:
        url = "https://www.animepahe.com/anime/%s" % player_match.group(1)

    anime_page = session.get(url)
    release_id = ID_RE.search(anime_page.text).group(1)

    fpd = get_session_page(session, "1", release_id)

    if fpd.get("last_page") == 1:
        yield from get_stream_urls_from_page(session, release_id, "1", check)
        return

    for page in page_minimization(predict_pages(fpd.get("total"), check)):
        yield from get_stream_urls_from_page(session, release_id, page, check)


def metadata_fetcher(session, url, match):
    player_match = PLAYER_RE.search(url)

    if player_match:
        url = "https://www.animepahe.com/anime/%s" % player_match.group(1)

    return {
        "titles": TITLES_REGEX.findall(session.get(url).text),
    }



import functools
from collections import defaultdict

API_ACCESS_TOKEN = "Basic BCoB9f4m4lSlo+fp05PjlwWcplxQXDT+N+1FfvsyoF41YSy8nH+kuJBQowYrVkiZq6PvTvjFEoQQvzJOt3pJZA=="
CLIENT_REFRESH_TOKEN = "qUwYw6UJ913jR/lWDnTEaTaDDDp2Ft1yvMjcrHTdetK6QboEpAzEsSqf7nvqcCAQsiiGc9dTROkpHgxLOcid5F7IGiXoUv8rbVe/Pnxy0v6//+3iWDKK9kZyltHcvkm6NshhsGq3tLfc6CHvmvZ//RXU9jepqlM/lFMpUq36LWpHvIV/ULZQLnmlfyNBNNiRL0TjKAd33aGfjZa1nYQV4qTHgTxEzc1hHJnUxVZ0Lq7PjeJGRp/ATMgWvActkfzk"


@functools.lru_cache()
def fetch_api_grant_token(session, api_endpoint):
    return "%(token_type)s %(access_token)s" % (
        session.post(
            api_endpoint + "auth/v1/token",
            data={
                "refresh_token": CLIENT_REFRESH_TOKEN,
                "grant_type": "refresh_token",
                "scope": "all",
            },
            headers={
                "Authorization": API_ACCESS_TOKEN,
            },
        ).json()
    )


def fetch_seasons(session, api_endpoint, media_id, *, predicate=None):
    media = session.get(
        api_endpoint + "content/v1/seasons",
        headers={
            "Authorization": fetch_api_grant_token(session, api_endpoint),
        },
        params={"channel_id": "crunchyroll", "id": media_id, "locale": "en-US"},
    ).json()

    episodes = defaultdict(list)

    for season in media.get("items", []):
        for episode in season.get("episodes", []):
            if predicate is None or predicate(episode):
                episodes[episode["episode_number"]].append(episode)

    return episodes


def get_media_streams(
    session,
    api_endpoint,
    media_id,
    *,
    channel_id="crunchyroll",
    locale="en-US",
    media_type="adaptive_hls"
):
    return session.get(
        api_endpoint + "videos/v1/streams",
        params={
            "channel_id": channel_id,
            "locale": locale,
            "id": media_id,
            "type": media_type,
        },
        headers={
            "Authorization": fetch_api_grant_token(session, api_endpoint),
        },
    ).json()



import logging
import os
import pathlib
import time

import httpx
import regex
import yarl
from tqdm import tqdm

from ...config import FFMPEG_EXECUTABLE, FFMPEG_HLS, FFMPEG_SUBMERGE
from .content_mt import mimetypes
from .ffmpeg import FFMPEG_EXTENSIONS, ffmpeg_download, has_ffmpeg, merge_subtitles
from .hls import HLS_STREAM_EXTENSIONS, hls_yield

if FFMPEG_EXECUTABLE:
    ffmpeg_executable = FFMPEG_EXECUTABLE

EXEMPT_EXTENSIONS = ["mpd"]
CONTENT_DISP_RE = regex.compile(r'filename=(?:"(.+?)"|([^;]+))')


def sanitize_filename(f):
    return "".join(" - " if _ in '<>:"/\\|?*' else _ for _ in f).strip()


def ext_from_filename(filename):
    position = filename.rfind(".")
    if position == -1:
        return ""
    return filename[position + 1 :]


def get_extension(url):
    url = yarl.URL(url)
    return ext_from_filename(url.name)


def guess_extension(content_type):
    if not content_type:
        return ""

    for _, cd, extension in mimetypes:
        if cd == content_type:
            return (extension or "").lstrip(".")


def ext_from_content_disposition(content_disposition):
    match = CONTENT_DISP_RE.search(content_disposition)

    if not match:
        return ""

    return ext_from_filename(match.group(1) or match.group(2))


def process_url(session, url, headers={}):
    """
    Get the extension, content size and range downloadability forehand.

    Returns:

    `str`, `int`, `bool`
    """
    response = session.head(url, headers=headers)
    response_headers = response.headers

    extension = (
        ext_from_content_disposition(
            response_headers.get("content-disposition", "") or ""
        )
        or guess_extension(response_headers.get("content-type", ""))
        or get_extension(url)
        or get_extension(str(response.url))
    ).lower()

    return (
        extension,
        int(response_headers.get("content-length") or 0),
        "bytes" in response_headers.get("accept-ranges", ""),
    )


def standard_download(
    session: httpx.Client,
    url: str,
    content_dir: pathlib.Path,
    outfile_name: str,
    extension: str,
    content_size: int,
    headers: dict = {},
    ranges=True,
    **opts
):
    file = "{}.{}".format(outfile_name, extension)

    logger = logging.getLogger("downloader/standard[{}]".format(file))
    out_path = content_dir / pathlib.Path(sanitize_filename(file))

    if not ranges:
        logger.critical(
            "Stream does not support ranged downloading; failed downloads cannot be continued."
        )

    with open(out_path, "ab") as outstream:
        downloaded = outstream.tell() if ranges else 0
        progress_bar = tqdm(
            desc="GET / {}".format(file),
            total=content_size,
            disable=opts.get("log_level", 20) > 20,
            initial=downloaded,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        )
        while content_size > downloaded:
            temporary_headers = headers.copy()
            if ranges:
                temporary_headers.update({"Ranges": "bytes={}-".format(downloaded)})
            try:
                with session.stream("GET", url, headers=headers) as http_stream:
                    http_stream.raise_for_status()

                    for chunks in http_stream.iter_bytes():
                        size = len(chunks)
                        outstream.write(chunks)
                        progress_bar.update(size)
                        downloaded += size
            except httpx.RequestError as e:
                if not ranges:
                    downloaded = 0
                    outstream.seek(0)
                    progress_bar.clear()
                else:
                    outstream.flush()
                logger.error("Downloading error due to {!r}, retrying.".format(e))
                time.sleep(opts.get("retry_timeout") or 5.0)

    progress_bar.close()


def hls_download(
    session: httpx.Client,
    url: str,
    content_dir: pathlib.Path,
    outfile_name: str,
    headers: dict = {},
    **opts
):

    sanitized = sanitize_filename(outfile_name)
    content_path = content_dir / "{}.ts".format(sanitized)
    index_holder = content_dir / "{}.partialts".format(sanitized)
    index = 1

    if index_holder.exists():
        with open(index_holder, "r") as ih:
            index = int(ih.read() or 1)

    sizes = []
    with open(content_path, "ab") as tsstream, open(index_holder, "w+") as istream:
        downloaded = tsstream.tell()
        if downloaded:
            sizes.extend([downloaded / index] * index)
        progress_bar = tqdm(
            desc="HLS GET / {}.ts".format(outfile_name),
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            initial=downloaded,
            disable=opts.get("log_level", 20) > 20,
        )

        for content in hls_yield(
            session,
            [{"stream_url": url, "headers": headers}],
            opts.get("preferred_quality") or 1080,
            opts.get("retry_timeout") or 5,
            continuation_index=index,
        ):
            stream = content.get("bytes")
            total = content.get("total")
            current = content.get("current")

            size = len(stream)
            sizes.append(size)

            mean = sum(sizes) / len(sizes)
            stddev = (sum(abs(mean - s) ** 2 for s in sizes) / len(sizes)) ** 0.5

            progress_bar.total = mean * total
            progress_bar.desc = "HLS GET / {}.ts [± {:.3f} MB]".format(
                outfile_name, stddev / (1024**2)
            )
            progress_bar.update(size)

            istream.write(str(current or 0 + 1))
            istream.seek(0)
            istream.flush()

            tsstream.write(stream)

    progress_bar.close()
    os.remove(index_holder)


def idm_download(url, headers, content_dir, outfile_name, extension, **opts):
    file = "{}.{}".format(outfile_name, extension)
    from .idmanlib import wait_until_download as idmdl

    idmdl(url, headers=headers or {}, download_folder=content_dir, filename=file)


def subautomatic(f):
    def __inner__(session, url, headers, content_dir, outfile_name, *args, **kwargs):
        logger = logging.getLogger("ffmpeg/submerge")

        subtitles = kwargs.pop("subtitles", [])

        callback = f(session, url, headers, content_dir, outfile_name, *args, **kwargs)

        if not subtitles:
            return

        if (not has_ffmpeg()) or not FFMPEG_SUBMERGE:
            logger.debug("{!r} will not be merged.".format(subtitles))
            for count, subtitle in enumerate(subtitles, 1):
                handle_download(
                    session,
                    subtitle,
                    headers=headers,
                    content_dir=content_dir,
                    outfile_name="{}_SUB_{}".format(outfile_name, count),
                )
            return callback

        extension, _, _ = process_url(session, url, headers)
        resolved_path = (
            content_dir / "{}.{}".format(outfile_name, extension)
        ).resolve()
        subout_path = (
            content_dir / "{} [CC].{}".format(outfile_name, extension)
        ).resolve()

        ffmpeg_returncode = merge_subtitles(
            resolved_path, subout_path, subtitles, log_level=kwargs.get("log_level", 20)
        )

        if ffmpeg_returncode:
            return (
                logger.warning(
                    "Got a non-zero return code {} from ffmpeg, the original file will not be erased in case of merge failure.".format(
                        ffmpeg_returncode
                    )
                )
                or callback
            )

        try:
            os.remove(resolved_path)
        except FileNotFoundError:
            pass

        try:
            os.rename(subout_path, resolved_path)
        except FileNotFoundError:
            pass

        return callback

    return __inner__


@subautomatic
def handle_download(
    session, url, headers, content_dir, outfile_name, idm=False, **opts
):

    extension, content_size, ranges = process_url(session, url, headers)

    if FFMPEG_HLS and (extension in FFMPEG_EXTENSIONS and has_ffmpeg()):
        return ffmpeg_download(url, headers, outfile_name, content_dir, **opts)

    if extension in EXEMPT_EXTENSIONS:
        raise Exception(
            "Download extension {!r} requires custom downloading which is not supported yet.".format(
                extension
            )
        )

    if extension in HLS_STREAM_EXTENSIONS:
        return hls_download(
            session, url, content_dir, outfile_name, headers or {}, **opts
        )

    if idm:
        return idm_download(
            url, headers or {}, content_dir.resolve(), outfile_name, extension, **opts
        )

    return standard_download(
        session,
        url,
        content_dir,
        outfile_name,
        extension,
        content_size,
        headers or {},
        ranges,
        **opts
    )



"""
Fun helpers for the cli.

Credits for the functions:
    create_random_titles: (https://www.ruggenberg.nl/titels.html)
"""

import logging
import os
from contextlib import suppress
from random import choice

import regex
import yarl

from ...__version__ import __core__
from ..http_client import client
from .constants import LABELS, LANGUAGE


terminal_columns = 0

with suppress(Exception):
    terminal_columns = os.get_terminal_size().columns


def line_chop(string: str, max_length, separators=[" ", "\n"]):
    if not string:
        return

    if len(string) <= max_length:
        yield string
        return

    sep, sep_index = max(
        ((_, string[:max_length].rfind(_)) for _ in separators), key=lambda x: x[1]
    )

    if sep_index == -1:
        sep, sep_index = "", max_length

    yield string[:sep_index]
    yield from line_chop(
        string[sep_index + len(sep) :], max_length, separators=separators
    )


def terminal_center(string: str, *, columns=terminal_columns):

    if not columns:
        return string

    def genexp():
        for line in string.splitlines():
            for piece in line_chop(line, columns):
                yield piece.center(columns)

    return "\n".join(genexp())


package_banner = terminal_center(
    """\
justfoolingaround/animdl - v{}
A highly efficient anime downloader and streamer\
""".format(
        __core__
    )
)


update_banner = terminal_center("Update available: {}")


def create_random_titles():

    adjs = LANGUAGE.get("adjective")
    noun = LANGUAGE.get("noun")

    return [
        "{}-{}".format(choice(adjs), choice(noun)),
        "the-{}-{}".format(choice(adjs), choice(noun)),
        "{}-{}".format(choice(noun), choice(noun)),
        "the-{}'s-{}".format(choice(noun), choice(noun)),
        "the-{}-of-the-{}".format(choice(noun), choice(noun)),
        "{}-in-the-{}".format(choice(noun), choice(noun)),
    ]


def to_stdout(message, caller="animdl", *, color_index=36):
    if caller:
        message = "[\x1b[{}m{}\x1b[39m] ".format(color_index, caller) + message
    return print(message)


def stream_judiciary(url):

    try:
        url = yarl.URL(url)
    except Exception:
        return "Unknown [URL Parsing error.]"

    return "{!r} from {}".format(url.name or "Unknown", LABELS.get(url.host, url.host))


def check_for_update(
    *,
    current=__core__,
    git_version_url="https://raw.githubusercontent.com/justfoolingaround/animdl/master/animdl/core/__version__.py"
):
    upstream_version = regex.search(
        r'__core__ = "(.*?)"', client.get(git_version_url).text
    ).group(1)
    return upstream_version == current, upstream_version


def bannerify(f):
    def internal(*args, **kwargs):
        quiet_state = kwargs.get("log_level")
        if quiet_state is not None:
            if quiet_state <= 20:
                print("\x1b[35m{}\x1b[39m".format(package_banner))
                latest, version = check_for_update()
                if not latest:
                    print("\x1b[36m{}\x1b[39m".format(update_banner.format(version)))
            logging.basicConfig(
                level=quiet_state,
                format="[\x1b[35m%(filename)s:%(lineno)d\x1b[39m - %(asctime)s - %(name)s: %(levelname)s] %(message)s",
                filename=kwargs.get("log_file"),
                filemode="a",
            )

            logger = logging.getLoggerClass()
            logger.FILE_STREAM = kwargs.get("log_file")

        return f(*args, **kwargs)

    return internal



import json as json_
import logging

import click

from ...config import DEFAULT_PROVIDER
from ..helpers import bannerify
from ..helpers.searcher import link
from ..http_client import client


@click.command(name="search", help="Search for an anime in the provider.")
@click.argument("query", required=True)
@click.option(
    "-p",
    "--provider",
    help="Provider to search in.",
    default=DEFAULT_PROVIDER,
    type=click.Choice(link.keys(), case_sensitive=False),
)
@click.option(
    "-ll", "--log-level", help="Set the integer log level.", type=int, default=20
)
@click.option(
    "--log-file",
    help="Set a log file to log everything to.",
    required=False,
)
@click.option(
    "-j",
    "--json",
    help="Output as json.",
    is_flag=True,
    flag_value=True,
)
@bannerify
def animdl_search(query, json, provider, **kwargs):
    logger = logging.getLogger("searcher")
    genexp = link.get(provider)(client, query)

    for count, search_data in enumerate(genexp, 1):
        if json:
            print(json_.dumps(search_data))
        else:
            logger.info("{0:02d}: {1[name]} {1[anime_url]}".format(count, search_data))



import json
from html import unescape

import lxml.html as htmlparser

OKRU = "https://ok.ru/"

QUALITY = {"mobile": 144, "lowest": 240, "low": 360, "sd": 480, "hd": 720}


def extract(session, url, **opts):
    response = session.get(url)

    if response.status_code >= 400:
        return []

    metadata = htmlparser.fromstring(response.text).cssselect(
        'div[data-module="OKVideo"]'
    )

    if not metadata:
        return []

    data_opts = json.loads(
        json.loads(unescape(metadata[0].get("data-options")))
        .get("flashvars", {})
        .get("metadata")
    )

    def fast_yield():
        for videos in data_opts.get("videos", []):
            yield {
                "quality": QUALITY.get(videos.get("name"), 1080),
                "stream_url": videos.get("url"),
                "headers": {"referer": OKRU},
            }

    return list(fast_yield()) + [
        {"stream_url": data_opts.get("hlsManifestUrl"), "headers": {"referer": OKRU}}
    ]



from functools import partial

import regex

from collections import defaultdict

from ....config import CRUNCHYROLL, KAMYROLL_API
from ..crunchyroll import metadata_fetcher
from .api import get_media_streams

REGEX = regex.compile(r"kamyroll://(?:.+?\.)+?crunchyroll\.com/([^&?/]+)")

MEDIA_ID_REGEX = regex.compile(r'\{"isAuthenticated":.+?"mediaId":"(.+?)".+?\}')

CRUNCHYROLL_EPISODES_REGEX = regex.compile(
    r"<crunchyroll:mediaId>(\d+)</crunchyroll:mediaId>"
    r".+?<crunchyroll:episodeTitle>(.+?)</crunchyroll:episodeTitle>"
    r".+?<crunchyroll:episodeNumber>(\d+)</crunchyroll:episodeNumber>",
    flags=regex.DOTALL,
)


def extract_streams(session, medias: "list"):

    collected_streams = []
    collected_subtitles = []

    for (media_id, title) in medias:
        streams = get_media_streams(session, KAMYROLL_API, media_id)

        collected_subtitles.extend(
            _["url"] for _ in streams.get("subtitles", []) if _["locale"] == "en-US"
        )

        for stream in streams.get("streams", []):
            if stream["url"] not in streams and not stream["hardsub_locale"]:
                collected_streams.append(
                    {
                        "stream_url": stream["url"],
                        "title": title,
                        **(
                            {"subtitle": collected_subtitles}
                            if collected_subtitles
                            else {}
                        ),
                    }
                )

    return collected_streams


def get_media_from_rss(session, rss_url):

    episodes = defaultdict(list)

    for match in list(CRUNCHYROLL_EPISODES_REGEX.finditer(session.get(rss_url).text))[
        ::-1
    ]:
        episodes[int(match.group(3))].append(match.group(1, 2))

    return episodes


def fetcher(session, url, check, match):
    rss_url = CRUNCHYROLL + match.group(1) + ".rss"

    for episode, medias in get_media_from_rss(
        session,
        rss_url,
    ).items():
        if check(int(episode)):
            yield partial(extract_streams, session, medias), episode
